import logging
import traceback
from types import MappingProxyType
from typing import Any

import fitz
from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.pandas import DataFrame
from pyspark.sql.functions import udf
from pyspark.sql.types import Row

from scaledp.params import (
    HasColumnValidator,
    HasInputCol,
    HasKeepInputData,
    HasOutputCol,
    HasPageCol,
    HasPathCol,
)
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document


class PdfDataToDocument(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasPathCol,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColumnValidator,
):
    """Extract text with coordinates from PDF file."""

    DEFAULT_PARAMS = MappingProxyType(
        {
            "inputCol": "content",
            "outputCol": "document",
            "pathCol": "path",
            "pageCol": "page",
            "keepInputData": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(PdfDataToDocument, self).__init__()
        self._setDefault(**self.DEFAULT_PARAMS)
        self._set(**kwargs)

    def transform_udf(self, input: Row, path: str) -> list[Document]:
        logging.info("Run Pdf Data to Text")
        try:
            doc = fitz.open("pdf", input)
            if len(doc) == 0:
                raise ValueError("Empty PDF document.")

            page = doc[0]
            words = page.get_text("words")

            boxes = []
            text_content = []

            for word in words:
                x0, y0, x1, y1, word_text, _, _, _ = word
                boxes.append(
                    Box(
                        x=int(x0),
                        y=int(y0),
                        width=int(x1 - x0),
                        height=int(y1 - y0),
                        text=word_text,
                        score=1.0,
                    ),
                )
                text_content.append(word_text)

            return Document(
                path=path,
                text=" ".join(text_content),
                type="pdf",
                bboxes=boxes,
            )

        except Exception:
            exception = traceback.format_exc()
            exception = (
                f"{self.uid}: Error during extracting text from "
                f"the PDF document: {exception}"
            )
            logging.warning(exception)
            return [
                Document(
                    path=path,
                    text="",
                    type="pdf",
                    bboxes=[],
                    exception=exception,
                ),
            ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        out_col = self.getOutputCol()
        input_col = self._validate(self.getInputCol(), dataset)
        path_col = dataset[self.getPathCol()]

        result = dataset.withColumn(out_col, udf(self.transform_udf, Document.get_schema())(
                    input_col,
                    path_col,
                ))
        if not self.getKeepInputData():
            result = result.drop(input_col)
        return result



import base64
import json
import logging
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

from scaledp.params import HasLLM, HasPrompt, HasSchema

from .BaseVisualExtractor import BaseVisualExtractor


class LLMVisualExtractor(BaseVisualExtractor, HasLLM, HasSchema, HasPrompt):

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "data",
            "keepInputData": True,
            "model": "gemini-1.5-flash",
            "apiBase": None,
            "apiKey": None,
            "numPartitions": 1,
            "pageCol": "page",
            "pathCol": "path",
            "prompt": """Please extract data from the scanned image as json.
         Date format is yyyy-mm-dd""",
            "systemPrompt": "You are data extractor from the scanned images.",
            "delay": 30,
            "maxRetry": 6,
            "propagateError": False,
            "temperature": 1.0,
            "schemaByPrompt": True,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(LLMVisualExtractor, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def call_extractor(self, images, params):
        from openai import RateLimitError

        client = self.getOIClient()
        results = []

        @retry(
            retry=retry_if_exception_type(RateLimitError),
            wait=wait_random_exponential(min=1, max=self.getDelay()),
            stop=stop_after_attempt(self.getMaxRetry()),
        )
        def completion_with_backoff(**kwargs: Any):
            logging.info("Calling LLM API")
            return client.beta.chat.completions.parse(**kwargs)

        kwargs = {}

        content = [{"type": "text", "text": params["prompt"]}]
        if self.getSchemaByPrompt():
            content.append(
                {
                    "type": "text",
                    "text": "Schema for the output json: "
                    + self.getSchema()
                    + " Always return valid json. Do not include schema to the output.",
                },
            )
        else:
            kwargs["response_format"] = self.getPaydanticSchema()

        for image in images:
            image_decoded = base64.b64encode(image.data).decode("utf-8")
            completion = completion_with_backoff(
                model=params["model"],
                messages=[
                    {
                        "role": "system",
                        "content": params["systemPrompt"],
                    },
                    {
                        "role": "user",
                        "content": [
                            *content,
                            *[
                                {
                                    "type": "image_url",
                                    "image_url": {
                                        "url": f"data:image/jpeg;base64,{image_decoded}",
                                    },
                                },
                            ],
                        ],
                    },
                ],
                temperature=self.getTemperature(),
                **kwargs,
            )

            result = (
                completion.choices[0]
                .message.content.replace("```json", "")
                .replace("```", "")
            )

            data = json.loads(result)
            results.append(
                self.getOutputClass()(
                    path=image.path,
                    json_data=json.dumps(data, indent=4, ensure_ascii=False),
                    type="LLMVisualExtractor",
                    data=data,
                    exception="",
                ),
            )
        return results



import itertools
from pathlib import Path
from typing import Any, ClassVar, List

import pandas as pd
from pyspark.ml import Transformer


class UserDefinedFunction:
    """
    User defined function in Python.

    .. versionadded:: 1.3

    Notes
    -----
    The constructor of this class is not supposed to be directly called.
    Use :meth:`pyspark.sql.functions.udf` or :meth:`pyspark.sql.functions.pandas_udf`
    to create this instance.

    """

    def __init__(
        self,
        func,
        returnType,
        name,
        evalType,
        deterministic: bool = True,
    ) -> None:
        if not callable(func):
            raise ValueError("Invalid function: not a function or callable")
        self.func = func
        self.returnType = returnType

    def __call__(self, *cols: Any) -> Any:
        cols = zip(*cols)
        return [self.func(*i) for i in cols]

    def _wrapped(self) -> Any:
        return self


def lit(value) -> Any:
    """Creates a :class:`Column` of literal value."""
    return itertools.repeat(value)


def _invoke_function(name: str, *args: Any) -> Any:
    if name == "lit":
        return lit(*args)
    raise ValueError("Invalid function name: %s" % name)


temp_functions = {}


def pathSparkFunctions(pyspark: Any) -> None:
    """Path Spark functions."""
    temp_functions["udf"] = pyspark.sql.udf.UserDefinedFunction
    pyspark.sql.udf.UserDefinedFunction = UserDefinedFunction
    temp_functions["invoke_function"] = pyspark.sql.functions._invoke_function
    pyspark.sql.functions._invoke_function = _invoke_function


def unpathSparkFunctions(pyspark: Any) -> None:
    """Unpath Spark functions."""
    pyspark.sql.udf.UserDefinedFunction = temp_functions["udf"]
    pyspark.sql.functions._invoke_function = temp_functions["invoke_function"]


class DatasetPd(pd.DataFrame):

    def withColumn(self, name, col) -> "DatasetPd":
        if name in self.columns:
            self[name] = col
        else:
            self.insert(0, name, col, True)
        return self

    def drop(self, col) -> "DatasetPd":
        return self

    def repartition(self, numPartitions) -> "DatasetPd":
        return self

    def coalesce(self, numPartitions) -> "DatasetPd":
        return self


class PandasPipeline:
    """PandasPipeline for call Spark ML Pipelines with Pandas DataFrame."""

    stages: ClassVar[List[Transformer]] = []

    def setStages(self, value) -> "PandasPipeline":
        self.stages = value
        return self

    def __init__(self, stages) -> None:
        self.setStages(stages)

    def fromFile(self, filename: str) -> Any:
        with Path.open(filename, "rb") as f:
            data = f.read()

        data = DatasetPd({"content": [data], "path": [filename], "resolution": [0]})

        for stage in self.stages:
            data = stage._transform(data)

        return data

    def fromBinary(self, data, filename="memory") -> Any:
        data = DatasetPd({"content": [data], "path": [filename], "resolution": [0]})
        for stage in self.stages:
            data = stage._transform(data)
        return data

    def fromPandas(self, data: pd.DataFrame) -> Any:
        data = DatasetPd(data)
        for stage in self.stages:
            data = stage._transform(data)
        return data

    def fromDict(self, data: dict) -> Any:
        data = DatasetPd(data)
        for stage in self.stages:
            data = stage._transform(data)
        return data



from dataclasses import dataclass
from typing import Any, Dict

from scaledp.utils.dataclass import map_dataclass_to_struct, register_type


@dataclass(order=True)
class Box:
    """Box object for represent bounding box data in Spark Dataframe."""

    text: str
    score: float
    x: int
    y: int
    width: int
    height: int

    def to_string(self) -> "Box":
        self.text = str(self.text)
        return self

    def json(self) -> Dict[str, Any]:
        return {"text": self.text}

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(Box)

    def scale(self, factor: float, padding: int = 0) -> "Box":
        return Box(
            text=self.text,
            score=self.score,
            x=int(self.x * factor) - padding,
            y=int(self.y * factor) - padding,
            width=int(self.width * factor) + padding,
            height=int(self.height * factor) + padding,
        )

    def shape(self, padding: int = 0) -> list[tuple[int, int]]:
        return [
            (self.x - padding, self.y - padding),
            (self.x + self.width + padding, self.y + self.height + padding),
        ]

    def bbox(self, padding: int = 0) -> list[int]:
        return [
            self.x - padding,
            self.y - padding,
            self.x + self.width + padding,
            self.y + self.height + padding,
        ]

    @staticmethod
    def from_bbox(box: list[int], label: str = "", score: float = 0):
        return Box(
            text=label,
            score=float(score),
            x=int(box[0]),
            y=int(box[1]),
            width=int(box[2] - box[0]),
            height=int(box[3] - box[1]),
        )


register_type(Box, Box.get_schema)



# Description: Schema Detector Output
from dataclasses import dataclass

from scaledp.utils.dataclass import map_dataclass_to_struct, register_type


@dataclass(order=True)
class ExtractorOutput:
    path: str
    data: str
    type: str
    exception: str = ""
    processing_time: float = 0.0

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(ExtractorOutput)


register_type(ExtractorOutput, ExtractorOutput.get_schema)



import gc
from pathlib import Path
from types import MappingProxyType
from typing import Any

from huggingface_hub import hf_hub_download
from pyspark import keyword_only

from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput

from ...enums import Device


class YoloDetector(BaseDetector, HasDevice, HasBatchSize):
    _model = None

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(YoloDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        if cls._model:
            return cls._model
        from ultralytics import YOLO

        model = params["model"]
        if not Path(model).is_file():
            model = hf_hub_download(repo_id=model, filename="best.pt")

        detector = YOLO(model)
        device = "cpu" if int(params["device"]) == Device.CPU.value else "cuda"

        cls._model = detector.to(device)
        return cls._model

    @classmethod
    def call_detector(cls, images, params):
        import torch

        detector = cls.get_model(params)

        results = detector(
            [image[0] for image in images],
            conf=params["scoreThreshold"],
            save_conf=True,
        )

        results_final = []
        for res, (_image, image_path) in zip(results, images):
            boxes = []
            for box in res.boxes:
                boxes.append(Box.from_bbox(box.xyxy[0]))
            results_final.append(
                DetectorOutput(path=image_path, type="yolo", bboxes=boxes),
            )

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results_final



import base64
import io
import logging
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

from scaledp.models.recognizers.BaseOcr import BaseOcr
from scaledp.params import HasLLM, HasPrompt
from scaledp.schemas.Document import Document


class LLMOcr(BaseOcr, HasLLM, HasPrompt):

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "text",
            "bypassCol": "",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "lang": ["eng"],
            "lineTolerance": 0,
            "keepFormatting": False,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "systemPrompt": "You are ocr.",
            "prompt": """Please extract text from the image.""",
            "model": "gemini-1.5-flash",
            "apiBase": None,
            "apiKey": None,
            "delay": 30,
            "maxRetry": 6,
            "propagateError": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(LLMOcr, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_ocr(cls, images, params):
        from openai import RateLimitError

        client = cls.getClient(params["apiKey"], params["apiBase"])
        results = []

        @retry(
            retry=retry_if_exception_type(RateLimitError),
            wait=wait_random_exponential(min=1, max=params["delay"]),
            stop=stop_after_attempt(params["maxRetry"]),
        )
        def completion_with_backoff(**kwargs: Any):
            logging.info("Calling LLM API")
            return client.beta.chat.completions.parse(**kwargs)

        for image, image_path in images:
            buff = io.BytesIO()
            image.save(buff, "png")
            image_decoded = base64.b64encode(buff.getvalue()).decode("utf-8")
            completion = completion_with_backoff(
                model=params["model"],
                messages=[
                    {
                        "role": "system",
                        "content": params["systemPrompt"],
                    },
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image/jpeg;base64,{image_decoded}",
                                },
                            },
                            {"type": "text", "text": params["prompt"]},
                        ],
                    },
                ],
            )

            results.append(
                Document(
                    path=image_path,
                    text=completion.choices[0].message.content,
                    type="text",
                    bboxes=[],
                ),
            )
        return results



import io
import logging
import traceback
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import udf

from scaledp.params import HasInputCol, HasOutputCol, HasResolution
from scaledp.schemas.PdfDocument import PdfDocument


class SingleImageToPdf(
    Transformer,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasOutputCol,
    HasInputCol,
    HasResolution,
):
    """Transform Image to PDF."""

    POINTS_PER_INCH = 72

    DEFAULT_PARAMS = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "pdf",
            "resolution": 0,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(SingleImageToPdf, self).__init__()
        self._setDefault(**self.DEFAULT_PARAMS)
        self._set(**kwargs)

    def transform_udf(self, image):
        try:
            if image.exception != "":
                return PdfDocument(
                    path=image.path,
                    data=bytes(),
                    exception=image.exception,
                )
            image_resolution = image.resolution
            if image_resolution == 0 or image_resolution is None:
                image_resolution = 300
            if self.getResolution() > 0:
                image_resolution = self.getResolution()

            width = image.width / image_resolution * self.POINTS_PER_INCH
            height = image.height / image_resolution * self.POINTS_PER_INCH

            import img2pdf

            a4inpt = (img2pdf.mm_to_pt(210), img2pdf.mm_to_pt(297))
            layout_fun = img2pdf.get_layout_fun(a4inpt)
            pdf_bytes = img2pdf.convert(io.BytesIO(image.data), layout_fun=layout_fun)
        except Exception:
            exception = traceback.format_exc()
            exception = f"SingleImageToPdf: {exception}, {image.exception}"
            logging.warning(exception)
            return PdfDocument(path=image.path, data=bytes(), exception=exception)

        return PdfDocument(
            path=image.path,
            data=pdf_bytes,
            width=width,
            height=height,
            exception="",
        )

    def _transform(self, dataset):

        output_col = self.getOutputCol()
        input_col = self.getInputCol()

        if input_col not in dataset.columns:
            raise ValueError(
                f"Input column '{input_col}' is not present in the DataFrame.",
            )

        input_col = dataset[self.getInputCol()]

        return dataset.withColumn(
            output_col,
            udf(self.transform_udf, PdfDocument.get_schema())(input_col),
        )



import logging
import traceback
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import lit, udf

from scaledp.enums import ImageType
from scaledp.params import (
    HasColumnValidator,
    HasDefaultEnum,
    HasImageType,
    HasInputCol,
    HasKeepInputData,
    HasOutputCol,
    HasPathCol,
    HasPropagateExc,
)
from scaledp.schemas.Image import Image


class ImageError(Exception):
    pass


class DataToImage(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasDefaultEnum,
    HasPathCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColumnValidator,
    HasPropagateExc,
):
    """Transform Binary Content to Image."""

    defaultParams = MappingProxyType(
        {
            "inputCol": "content",
            "outputCol": "image",
            "pathCol": "path",
            "keepInputData": False,
            "imageType": ImageType.FILE,
            "propagateError": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(DataToImage, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    def process(self, input, path, resolution):
        return Image.from_binary(
            input,
            path,
            self.getImageType(),
            resolution=resolution,
        )

    def transform_udf(self, input, path, resolution):
        try:
            return self.process(input, path, resolution)
        except Exception as ex:
            exception = traceback.format_exc()
            exception = f"DataToImage: {exception}"
            logging.warning(exception)
            if self.getPropagateError():
                raise ImageError(exception) from ex
            return Image(path, self.getImageType(), data=bytes(), exception=exception)

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        input_col = self._validate(self.getInputCol(), dataset)
        path_col = self._validate(self.getPathCol(), dataset)
        resolution = (
            dataset["resolution"] if "resolution" in dataset.columns else lit(0)
        )
        result = dataset.withColumn(
            out_col,
            udf(self.transform_udf, Image.get_schema())(
                input_col,
                path_col,
                resolution,
            ),
        )
        if not self.getKeepInputData():
            result = result.drop(input_col)
        return result
